{$A-,B-,D+,E+,F+,G-,I-,L+,N+,O-,P-,Q-,R-,S-,T-,V-,X+}
Unit StrCalc;
interface
var
 ErrorCode : Byte;
 function CalcStr(SStr: String) : String;
 procedure CalcStrInit;
 procedure CalcStrDone;

implementation
uses
 GetId, RevPos, TpString, Jump, Objects, Memory;
type
 CharSet = Set of Char;
 St9 = String[9];

 PVarItem = ^TVarItem;
 TVarItem = record
  Name : Pstring;
  Value : Double;
 end;

 PVarColl = ^TVarColl;
 TVarColl = object(TSortedCollection)
  procedure FreeItem(Item: Pointer); virtual;
  function Compare(Key1, Key2: Pointer): Integer; virtual;
 end;

const
 OpCount = 16;
 Operators : Array [0 .. OpCount] of String[2] =
  ('E-','E+','E','^','/','*','%','-','+','>=','<=','>','<','<>','=','&','|');
 IdChars : Set Of Char = ['$','0'..'9','_','.','A'..'Z','a'..'z'];
 TerminalChars : Set Of Char = ['$','-','.','_',' ','0'..'9','.','A'..'Z','a'..'z'];
 Digits : Set Of Char = ['0'..'9'];
 FnCount = 16;
 Functions : Array[0..FnCount] of String[7] =
 ('ARCCOS(', 'ARCSIN(', 'ARCTG(', 'SING(', 'SIN(', 'COSG(','COS(', 'CTGG(',
 'CTG(', 'TGG(', 'TG(',  'SQRT(',   'LG(',    'LN(',  'EXP10(', 'EXP(',
 'ABS(' );

 VarColl : PVarColl = Nil;
var
 ErrorStr : String[60];
 ExitSave : Procedure;


procedure TVarColl.FreeItem(Item: Pointer);
begin
 if Item <> Nil then
  begin
   if PVarItem(Item)^.Name <> Nil then  DisposeStr(PVarItem(Item)^.Name);
   FreeMem(Item, SizeOf(TVarItem));
  end;
end;

function TVarColl.Compare(Key1, Key2: Pointer): Integer;
begin
 if PVarItem(Key1)^.Name^ = PVarItem(Key2)^.Name^ then Compare := 0
 else if PVarItem(Key1)^.Name^ > PVarItem(Key2)^.Name^ then Compare := 1
      else Compare := -1
end;

procedure Error(Code : Byte);
begin
 ErrorCode  := Code;
 case Code of
  200 : ErrorStr := 'Division by zero';
  205 : ErrorStr := 'Floating point overflow';
  206 : ErrorStr := 'Floating point underflow';
  207 : ErrorStr := 'Invalid floating point operation';
  5 : ErrorStr := 'Syntax error';
  6..8 : ErrorStr := 'Invalid brackets';
  10 :  ErrorStr := 'Invalid argument';
  13 :  ErrorStr := 'String too long';
 else ErrorStr := 'Internal error. Contact the developers';
end {case};
 LongJump;
end;

procedure CalcExitProc; far;
begin
 if (ErrorAddr <> Nil) and (ExitCode in [200, 205 .. 207]) then
 Error(ExitCode)
 else ExitSave;
end;

function PosWord(const SubStr,Str: String;
                 const TerminalChars : CharSet) : Byte;
 var
  Left, Right : Boolean;
  Ps, Ps1, SS0, S0 : Byte;
 begin
  PosWord := 0;
  Ps1 := 0;
  SS0 := Byte(SubStr[0]);
  S0 := Byte(Str[0]);
  repeat
   Ps := FIPos(SubStr,Str,Ps1);
   if Ps = 0 then Exit;
   Left := (Ps = 1) or (Str[Ps-1] in TerminalChars);
   Right := (Ps + SS0 - 1 = S0) or (Str[Ps + SS0] in TerminalChars);
   Ps1 := Ps;
  until Left and Right;
  PosWord := Ps;
 end;

function GetKeyAddress(const S : String) : Pointer;
var
 J : Integer;
 PStr : PString;
 Item : PVarItem;
begin
 if VarColl <> Nil then
  begin
   PStr := @S;
   if not VarColl^.Search(@PStr,J) then
    begin
     Item := MemAlloc(SizeOf(TVarItem));
     Item^.Name := NewStr(S);
     Item^.Value := 0;
     VarColl^.AtInsert(J, Item);
    end;
    GetKeyAddress := Addr(PVarItem(VarColl^.At(J))^.Value)
   end
 else GetKeyAddress := Nil;
end;

function ParseKey(const S:String) : Double;
var
 J : Integer;
 PStr : PString;
 D : Double;
 L : Longint;
 Code : Integer;
begin

 PStr := @S;
 if (VarColl <> Nil) and VarColl^.Search(@PStr,J) then
  begin
   ParseKey := PVarItem(VarColl^.At(J))^.Value;
   Exit;
  end;

 Val(S, L, Code);
 if Code = 0 then
  begin
   ParseKey := L;
   Exit;
  end;

 Val(S, D, Code);
 if Code = 0 then
  begin
   ParseKey := D;
   Exit;
  end;

 if S = 'PI' then
  begin
   ParseKey := Pi;
   Exit;
  end;
 Error(5);
end;

function StrI(D : Double) : String;
var
 Code : Integer;
 SS : String;
begin
 Str(D:0:0,SS);
 StrI := SS;
end;

function StrD(D : Double) : String;
var
 Code : Integer;
 SS : String;
begin
 Str(D:0:10,SS);
 StrD := SS;
end;

{('E-','E+','E','^','/','*','%','-','+','>=','<=','>','<','<>','=','&','|');}
function CallOp(Right,Left : double; Op : Byte) : String;
var
 R : Double;
 B,BB : Byte;
 Code : Integer;
 S : String;
begin
 case Op of
  0 : {E-} Val(StrD(Left) + 'E-' + StrI(Right), R, Code);
  1,  {E+}
  2 : {E}  Val(StrD(Left) + 'E' + StrI(Right), R, Code);
  3 : {^}  R := Exp(Right * Ln(Left));
  4 : {/}  R := Left / Right;
  5 : {*}  R := Left * Right;
  6 : {%}  R := Trunc(Left / Right);
  7 : {-}  R := Left - Right;
  8 : {+}  R := Left + Right;
  9 : {>=} begin
            B := Byte(Left >= Right);
            R := B;
           end;
  10: {<=} begin
            B := Byte(Left <= Right);
            R := B;
           end;
  11:  {>} begin
            B := Byte(Left > Right);
            R := B;
           end;
  12: {<}  begin
            B := Byte(Left < Right);
            R := B;
           end;
  13: {<>}  begin
            B := Byte(Left <> Right);
            R := B;
           end;
  14 : {=}  begin
            B := Byte(Left = Right);
            R := B;
           end;
  15: {&}  begin
            BB := Trunc(Left); B := Trunc(Right);
            R := byte(boolean(B) and boolean(BB));
           end;
  16: {|}  begin
            BB := Trunc(Left); B := Trunc(Right);
            R := byte(boolean(B) or boolean(BB));
           end;
 end; {case}
{ if Op > 8 then Str(R:0:0,S)
 else} Str(R:0:10,S);
 {CallOp := DelChar(' ',S);}
 CallOp := S;
end;

procedure CompileStr(var SStr : String);
var
 OpIndex,Ps,Lps,Rps : Byte;
 Code : Integer;
 Result, RResult, LResult : Double;
 LSubStr, RSubStr : String;
 Minus : Boolean;
begin
if SStr = '' then Exit;
Ps := 0;
Minus := False;
for OpIndex := 0 to OpCount do
  repeat
   if OpIndex < 3 then Ps := PosWord(Operators[OpIndex],SStr,Digits)
   else Ps := PosWord(Operators[OpIndex],SStr,TerminalChars);
   if Ps <> 0  then
    begin
     Lps := Ps;
     LSubStr := GetLeftId(SStr,IdChars,LPs);
     if LSubStr <> '' then
      begin
       if Pos(LSubStr,SStr) = 0 then  Error(5); {Syntax Error}
       LResult := ParseKey(LSubStr);
       if Minus then
        begin
         LResult := - LResult;
         Minus := False;
        end;
      end
     else
      begin
       if Operators[OpIndex] <> '-' then Error(5); {Syntax Error}
       Delete(SStr,Ps,1);
       Minus := True;
       Continue;
      end;
     RPs := Ps  + Length(Operators[OpIndex]) - 1;
     if SStr[RPs+1] = '-' then
      begin
       Inc(RPs);
       RSubStr := '-' + GetRightId(SStr, IdChars, RPs);
      end
     else RSubStr := GetRightId(SStr,IdChars,RPs);
     if RSubStr = '' then Error(14); {Syntax Error}
     if Pos(RSubStr,SStr) = 0 then Error(5); {Syntax Error}
     RResult := ParseKey(RSubStr);

     LSubStr := CallOp(RResult,LResult,OpIndex);
     if Pos('E',LSubStr) <> 0 then Error(205);
     Delete(SStr,LPs,RPs-LPs + 1);
     Insert(LSubStr,SStr,LPs);
     if Word(Length(SStr)) + Length(LSubStr) > 256 then Error(13);
     Val(SStr,LResult,Code);
     if Code = 0 then Break;
    end;
  until Ps = 0;
  if Minus then SStr := '-' + SStr;
  Str(ParseKey(SStr):0:10,SStr);
end;

Procedure BrackMax(Var Str : String);
 var
  Index,OBr,Cbr : Byte;
  SubStr : String;
  begin
   OBr:=0; CBr:=0; Index:=1;
   while (CBr = 0) and (Index <= Length(Str)) do
    begin
     if Str[Index] = '(' then OBr := Index;
     if Str[Index] = ')' then CBr := Index+1;
     Inc(Index);
    end;
    if (CBr <> 0) and (OBr <> 0) and (CBr > OBr) then
       begin
         if (CBr <= Length(Str)) and (Str[Cbr] in IdChars) then Error(5);
         if (Obr > 1) and (Str[Obr-1] in IdChars) then Error(5);
         SubStr := Copy(Str,OBr,CBr-OBr);
         Delete(SubStr,1,1);
         Dec(SubStr[0]);
         CompileStr(SubStr);
         Delete(Str,OBr,CBr-OBr);
         if Word(Length(Str)) + Length(SubStr) > 256 then Error(13);
         Insert(SubStr,Str,Obr);
         BrackMax(Str);
       end
    else if (CBr = 0) and (OBr = 0) then CompileStr(Str)
         else if CBr < OBr then Error(6)  {() Error}
              else if CBr <> 0 then Error(7) {() Error}
                   else Error(8); {() Error}
  end;

 {('ARCCOS(', 'ARCSIN(', 'ARCTG(', 'SING(', 'SIN(', 'COSG(','COS(', 'CTGG(',
 'CTG(', 'TGG(', 'TG(',  'SQRT(',   'LG(',    'LN(',  'EXP10(', 'EXP(',
 'ABS(' );}
 procedure CallFn(FnIdx : Byte; var Arg : String);
 var
  R, FR : Double;
  Code : Integer;
 begin
  Val(Arg, R, Code);
  if Code <> 0 then Error(10);{  㬥 㭪樨}
  case FnIdx of
   0: {ARCCOS}    FR := ArcTan(Sqrt(1-Sqr(R))/R);
   1: {ARCSIN}    FR := ArcTan(R/Sqrt(1-Sqr(R)));
   2: {ARCTG}     FR := ArcTan(R);
   3: {SING}      FR := Sin(Pi*R/180);
   4: {SIN}       FR := Sin(R);
   5: {COSG}      FR := Cos(Pi*R/180);
   6: {COS}       FR := Cos(R);
   7: {CTGG}      FR := Cos(Pi*R/180)/Sin(Pi*R/180);
   8: {CTG}       FR := Cos(R)/Sin(R);
   9: {TGG}       FR := Sin(Pi*R/180)/Cos(Pi*R/180);
   10:{TG}        FR := Sin(R)/Cos(R);
   11:{SQRT}      FR := Sqrt(R);
   12:{LG}        FR := Ln(R)/Ln(10);
   13:{LN}        FR := Ln(R);
   14:{EXP10}     FR := Exp(R*ln(10));
   15:{EXP}       FR := Exp(R);
   16:{ABS}       FR := Abs(R);
  end;
  Str(FR:0:10,Arg);
 end;

 Procedure CompileFn(Var Str : String);
 var
  FnIndex, FnMIndex, FnPos, Ps,
  FnMPos, OBr, CBr, Args, N: Byte;
  SubStr,RSubStr : String[128];
 begin
  SetJump;
  repeat
   {    㭪}
   FnMpos := 0;
   for FnIndex := 0 to FnCount do
    begin
     FnPos := RPos(Functions[FnIndex],Str);
     if (FnPos > 1) and (Str[FnPos-1] in IdChars) then Continue;
     if FnPos > FnMPos then
      begin
       FnMPos := FnPos;
       FnMIndex := FnIndex;
      end;
    end;
   if FnMPos <> 0 then
    begin
     if (FnMPos > 1) and (Str[FnMPos-1] in IdChars) then Error(5);
     {஢塞,  । 㭪樥  ᫮}
     OBr := FIPos('(',Str,FnMPos);
     if OBr = 0 then Error(7); { "(" expected}
     CBr := findCBr(Str,OBr);
     if CBr = 0 then Error(6);
     if (CBr < Length(Str)) and (Str[Cbr+1] in IdChars) then Error(5);
     {஢塞,  ᫥ 㭪樨  ᫮}
     SubStr := Copy(Str,Obr + 1,Cbr - Obr-1);
     BrackMax(SubStr);
     CallFn(FnMIndex,SubStr);
     Delete(Str,FnMPos,CBr-FnMPos + 1);
     if Word(Length(Str)) + Length(SubStr) > 256 then Error(13);
     Insert(SubStr,Str,FnMPos);
    end;
  until FnMPos = 0;
  BrackMax(Str);
 end;

function CalcStr(SStr: String) : String;
var
 Ps : Byte;
 Code : Integer;
 Ptr : ^Double;
 D : Double;
 SubStr : String;
begin
 Ptr := Nil;
 ErrorCode := 0;
 Pointer(@ExitSave^) := ExitProc;
 ExitProc := @CalcExitProc;
 SStr := StUpCase(DelChar(' ',SStr));
{ Ps := PosWord(':=',SStr,TerminalChars);}
 Ps := Pos(':=',SStr);
 if Ps <> 0 then
  begin
    SubStr := Trim(Copy(SStr,1,Ps-1));
    if SubStr <> ''then
     begin
      SStr := Copy(SStr,Ps + 2,255);
      CompileFn(SStr);
      Ptr := GetKeyAddress(SubStr);
     end;
   end
 else CompileFn(SStr);
 ExitProc := Pointer(@ExitSave^);
 if ErrorCode <> 0 then CalcStr := ErrorStr
 else CalcStr := SStr;
 if Ptr <> Nil then
  begin
   Val(SStr, D, Code);
   if Code = 0 then Ptr^ := D;
  end;
end;

procedure CalcStrInit;
begin
 VarColl := New(PVarColl, Init(10,10));
end;

procedure CalcStrDone;
begin
 Dispose(VarColl,Done);
 VarColl := Nil;
end;

end.


